#include "itkImage.h"
#include "itkImageFileWriter.h"
#include "itkBSplineTransform.h"
#include "itkLandmarkBasedTransformInitializer.h"
#include "itkResampleImageFilter.h"
 
const unsigned int Dimension = 3;
typedef unsigned char                      PixelType;
typedef itk::Image< PixelType, Dimension > ImageType;

void CreateImage( const char *fileName, ImageType::Pointer image,
  ImageType::IndexType objectBegin, ImageType::IndexType objectEnd, PixelType objectValue );

int main(int argc, char * argv[])
{
  ImageType::Pointer fixedImage = ImageType::New();
  ImageType::IndexType fixedObjectBegin, fixedObjectEnd;
  fixedObjectBegin.Fill( 10 );
  fixedObjectEnd.Fill( 20 );
  CreateImage( "fixedImage.nii", fixedImage, fixedObjectBegin, fixedObjectEnd, 255 );
 
  ImageType::Pointer movingImage = ImageType::New();
  ImageType::IndexType movingObjectBegin, movingObjectEnd;
  movingObjectBegin.Fill( 70 );
  movingObjectEnd.Fill( 80 );
  CreateImage( "movingImage.nii", movingImage, movingObjectBegin, movingObjectEnd, 127 );

  typedef itk::BSplineTransform< double, Dimension, Dimension > BSplineTransformType;
  typedef itk::LandmarkBasedTransformInitializer< BSplineTransformType, ImageType, ImageType > 
      LandmarkBasedTransformInitializerType;
  typedef LandmarkBasedTransformInitializerType::LandmarkPointContainer LandmarkContainerType;
  typedef LandmarkBasedTransformInitializerType::LandmarkPointType      LandmarkPointType;
  
  ImageType::IndexType  fixedIndex, movingIndex;
  LandmarkPointType     fixedPoint, movingPoint;
  LandmarkContainerType fixedLandmarks, movingLandmarks;

  fixedIndex[0] = fixedObjectBegin[0];
  fixedIndex[1] = fixedObjectBegin[1];
  fixedIndex[2] = fixedObjectBegin[2];
  movingIndex[0] = movingObjectBegin[0];
  movingIndex[1] = movingObjectBegin[1];
  movingIndex[2] = movingObjectBegin[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectEnd[0];
  fixedIndex[1] = fixedObjectBegin[1];
  fixedIndex[2] = fixedObjectBegin[2];
  movingIndex[0] = movingObjectEnd[0];
  movingIndex[1] = movingObjectBegin[1];
  movingIndex[2] = movingObjectBegin[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectBegin[0];
  fixedIndex[1] = fixedObjectEnd[1];
  fixedIndex[2] = fixedObjectBegin[2];
  movingIndex[0] = movingObjectBegin[0];
  movingIndex[1] = movingObjectEnd[1];
  movingIndex[2] = movingObjectBegin[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectEnd[0];
  fixedIndex[1] = fixedObjectEnd[1];
  fixedIndex[2] = fixedObjectBegin[2];
  movingIndex[0] = movingObjectEnd[0];
  movingIndex[1] = movingObjectEnd[1];
  movingIndex[2] = movingObjectBegin[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectBegin[0];
  fixedIndex[1] = fixedObjectBegin[1];
  fixedIndex[2] = fixedObjectEnd[2];
  movingIndex[0] = movingObjectBegin[0];
  movingIndex[1] = movingObjectBegin[1];
  movingIndex[2] = movingObjectEnd[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectEnd[0];
  fixedIndex[1] = fixedObjectBegin[1];
  fixedIndex[2] = fixedObjectEnd[2];
  movingIndex[0] = movingObjectEnd[0];
  movingIndex[1] = movingObjectBegin[1];
  movingIndex[2] = movingObjectEnd[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectBegin[0];
  fixedIndex[1] = fixedObjectEnd[1];
  fixedIndex[2] = fixedObjectEnd[2];
  movingIndex[0] = movingObjectBegin[0];
  movingIndex[1] = movingObjectEnd[1];
  movingIndex[2] = movingObjectEnd[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );

  fixedIndex[0] = fixedObjectEnd[0];
  fixedIndex[1] = fixedObjectEnd[1];
  fixedIndex[2] = fixedObjectEnd[2];
  movingIndex[0] = movingObjectEnd[0];
  movingIndex[1] = movingObjectEnd[1];
  movingIndex[2] = movingObjectEnd[2];
  fixedImage->TransformIndexToPhysicalPoint( fixedIndex, fixedPoint );
  fixedLandmarks.push_back( fixedPoint );
  movingImage->TransformIndexToPhysicalPoint( movingIndex, movingPoint );
  movingLandmarks.push_back( movingPoint );
  
  BSplineTransformType::Pointer transform = BSplineTransformType::New();
  LandmarkBasedTransformInitializerType::Pointer landmarkBasedTransformInitializer =
    LandmarkBasedTransformInitializerType::New();
  transform->SetIdentity();
  landmarkBasedTransformInitializer->SetFixedLandmarks( fixedLandmarks );
  landmarkBasedTransformInitializer->SetMovingLandmarks( movingLandmarks );
  landmarkBasedTransformInitializer->SetTransform( transform );
  landmarkBasedTransformInitializer->SetBSplineNumberOfControlPoints( 5 );
  landmarkBasedTransformInitializer->SetReferenceImage( fixedImage );
  landmarkBasedTransformInitializer->InitializeTransform(); // it crashes here

  typedef itk::ResampleImageFilter< ImageType, ImageType, double > ResampleFilterType;
  ResampleFilterType::Pointer resampleFilter = ResampleFilterType::New();
  resampleFilter->SetInput( movingImage );
  resampleFilter->SetTransform( transform );
  resampleFilter->SetSize( fixedImage->GetLargestPossibleRegion().GetSize() );
  resampleFilter->SetOutputOrigin( fixedImage->GetOrigin() );
  resampleFilter->SetOutputSpacing( fixedImage->GetSpacing() );
  resampleFilter->SetOutputDirection( fixedImage->GetDirection() );
  resampleFilter->SetDefaultPixelValue( 63 );
  resampleFilter->GetOutput();
 
  typedef itk::ImageFileWriter< ImageType > WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetInput( resampleFilter->GetOutput() );
  writer->SetFileName( "outputImage.nii" );
  writer->Update();
 
  return EXIT_SUCCESS;
}

void CreateImage( const char *fileName, ImageType::Pointer image,
  ImageType::IndexType objectBegin, ImageType::IndexType objectEnd, PixelType objectValue )
{
  ImageType::IndexType start;
  start.Fill( 0 );
 
  ImageType::SizeType size;
  size.Fill( 128 );
 
  ImageType::RegionType region;
  region.SetSize( size );
  region.SetIndex( start );
 
  ImageType::PointType origin;
  origin.Fill( 0 );

  ImageType::SpacingType spacing;
  spacing.Fill( 1 );

  // It seems that the direction matrix is the source of the problem. If it contains negative
  // values (as it is for many non-oblique orientations), the program raises an exception
  // (segmentation fault in ITK <= v4.8.0).
  ImageType::DirectionType direction;
  direction[0][0] = 1; direction[0][1] = 0; direction[0][2] = 0;
  direction[1][0] = 0; direction[1][1] = 1; direction[1][2] = 0;
  direction[2][0] = 0; direction[2][1] = 0; direction[2][2] = -1;

  image->SetRegions( region );
  image->SetOrigin( origin );
  image->SetSpacing( spacing );
  image->SetDirection( direction );
  image->Allocate();
  image->FillBuffer( 0 );
 
  itk::ImageRegionIterator< ImageType > iterator( image, region );
 
  while( !iterator.IsAtEnd() )
    {
    if( iterator.GetIndex()[0] > objectBegin[0] && iterator.GetIndex()[0] < objectEnd[0] &&
        iterator.GetIndex()[1] > objectBegin[1] && iterator.GetIndex()[1] < objectEnd[1] &&
        iterator.GetIndex()[2] > objectBegin[2] && iterator.GetIndex()[2] < objectEnd[2] )
      {
      iterator.Set( 255 );
      }
    ++iterator;
    }
 
  typedef itk::ImageFileWriter< ImageType > WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetInput( image );
  writer->SetFileName( fileName );
  writer->Update();
}
